use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007006
use Test::Spelling 0.17;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( examples lib script t xt ) );
__DATA__
Annotation
Applicator
Content
Core
Document
Error
Etheridge
FormatAnnotation
FormatAssertion
JSON
Karen
MetaData
Modern
Result
ResultNode
Schema
Unevaluated
Utilities
Validation
Vocabulary
ether
irc
json
lib
script
